# CheapBombTimer release 7 by David Bowland
# ./addons/eventscripts/cheapbombtimer/cheapbombtimer.py

# >>> To configure this addon please see cheapbombtimer.cfg <<<

"""
Displays bomb time at specified intervals and emits a countdown voice from the bomb. Allows custom sounds but no client downloads required for default sounds. Also compatible with slow motion! Requires only ES 2.0+
"""


import es
import langlib
import os.path
import playerlib
import random

import psyco
psyco.full()


info = es.AddonInfo()
info.name     = 'CheapBombTimer'
info.version  = '7'
info.url      = 'http://addons.eventscripts.com/addons/view/cheapbombtimer'
info.basename = 'cheapbombtimer'
info.author   = 'SuperDave'


class BombSoundManager:
   class BombSound:
      sound    = {}
      triggers = []

      def __init__(self, sound):
         """Stores a sorted list of sound times"""
         self.sound       = sound
         self.triggers[:] = sorted(sound, reverse=True)

      def getSound(self, time):
         """Returns the sound of the first sound time if the current time is less, otherwise returns an empty string"""
         if self.triggers:
            if time <= self.triggers[0]:
               return_sound = self.sound[self.triggers[0]]
               self.triggers.remove(self.triggers[0])
               return return_sound

         return ''

   sounds     = {}
   current_id = ''
   sound_path = es.ServerVar('eventscripts_gamedir') + '/sound/'

   def __init__(self):
      self.sounds = {}

   def addSound(self, id, time, sound):
      """Adds the sound to the sounds dictionary by sound id and time"""
      if not self.sounds.has_key(id):
         self.sounds[id] = {}

      self.sounds[id][time] = sound

   def removeSound(self, id):
      """Removes sound id from sounds dictionary"""
      if self.sounds.has_key(id):
         del self.sounds[id]

   def markDownloads(self):
      """Loops through all sounds and marks those that exist on disk as downloadable"""
      for id in self.sounds:
         loop_sound = self.sounds[id]
         for sound in filter(lambda x: os.path.isfile(self.sound_path + loop_sound[x]), loop_sound):
            es.stringtable('downloadables', 'sound/' + sound)

   def nextSound(self):
      """
      Returns a BombSound instance with sounds from a random sound id
      Returns None if no sound ids
      """
      if self.sounds:
         choices         = filter(lambda x: x <> self.current_id, self.sounds)
         self.current_id = random.choice(choices if choices else self.sounds.keys())

         return self.BombSound(self.sounds[self.current_id])

      self.current_id = ''

      return None

bombsound = BombSoundManager()


class BombTextManager:
   class BombText:
      announcements = []

      def __init__(self, announce_time):
         """Stores a sorted version of the passed set in a list"""
         self.announcements = sorted(announce_time, reverse=True)

      def shouldAnnounce(self, time):
         """Returns the first announce time if the current time is less, otherwise returns 0"""
         if (time <= self.announcements[0]) if self.announcements else False:
            announce_time = self.announcements[0]
            self.announcements.remove(announce_time)

            return announce_time

         return 0

   announce_time = set()

   def __init__(self):
      self.announce_time = set()

   def addAnnouncements(self, times):
      """Adds all times in the passed object to the announce times"""
      self.announce_time = self.announce_time.union(times)

   def removeAnnouncements(self, times):
      """Removes all times in the passed object from the announce times"""
      self.announce_time = self.announce_time.difference(times)

   def getBombText(self):
      """Returns a BombText instance with the current announce times"""
      return self.BombText(self.announce_time)

bombtext = BombTextManager()


class BombTimerManager:
   tick_total = 0
   start_time = 0
   bps        = 0
   bomb_sound = None
   bomb_text  = None

   def __init__(self):
      self.resetTicks()

   def bombPlanted(self):
      """
      Sets instance attributes to reflect the new bomb
      """
      self.tick_total = 0
      self.start_time = int(es.ServerVar('mp_c4timer'))
      self.bps        = int(es.ServerVar('cheapbombtimer_bps')) / 10.
      self.bomb_sound = bombsound.nextSound()
      self.bomb_text  = bombtext.getBombText()

   def tick(self, bomb_index):
      """
      Increases the tick count and calculates the bomb time
      Checks for and plays bomb sound
      Checks for and announces bomb text
      """
      if not self.start_time: return

      self.tick_total += 1
      bomb_time        = round(self.start_time - self.tick_total / self.bps, 2)

      if self.bomb_sound:
         sound       = self.bomb_sound.getSound(bomb_time)
         attenuation = (100. - float(es.ServerVar('cheapbombtimer_sound_percent'))) / 100.
         if sound and attenuation:
            es.emitsound('entity', bomb_index, sound, 1.0, attenuation)

      announce_time = self.bomb_text.shouldAnnounce(bomb_time)
      if announce_time:
         chat   = int(es.ServerVar('cheapbombtimer_announce_chat'))
         center = int(es.ServerVar('cheapbombtimer_announce_center'))
         hud    = int(es.ServerVar('cheapbombtimer_announce_hud'))

         if chat or center or hud:
            tokens = {'time':announce_time}
            for userid in es.getUseridList():
               text = func_lang_text('bomb time', tokens, playerlib.getPlayer(userid).get('lang'))

               if chat:
                  es.tell(userid, '#multi', text)

               if center:
                  es.centertell(userid, self.__removeTags(text))

               if hud:
                  self._hudHint(userid, self.__removeTags(text))

   def bombExploded(self):
      """Checks the counted number of ticks and elapsed time against the BPS cvar"""
      if self.start_time and not int(es.ServerVar('cheapbombtimer_bps_skipcheck')):
         new_bps = int(round((self.tick_total + 1) * 10. / self.start_time))

         if new_bps <> self.bps * 10.:
            es.centermsg('cheapbombtimer_bps incorrectly configured!')
            es.msg('cheapbombtimer_bps incorrectly configured!')
            es.msg('You should change your cheapbombtimer_bps value to %s' % new_bps)
            es.msg('This setting can found in your cheapbombtimer.cfg')
            es.ServerVar('cheapbombtimer_bps').set(new_bps)

   def resetTicks(self):
      """Resets the number of bomb ticks and the skipcheck cvar"""
      self.start_time = 0
      es.ServerVar('cheapbombtimer_bps_skipcheck').set(0)

   def _hudHint(self, userid, message):
      """Sends the player a HUD hint"""
      es.usermsg('create', 'healmates_hud', 'HintText')
      es.usermsg('write', 'short', 'healmates_hud', -1)
      es.usermsg('write', 'string', 'healmates_hud', message)
      es.usermsg('send', 'healmates_hud', userid)
      es.usermsg('delete', 'healmates_hud')

   def __removeTags(self, text):
      """Removes #lightgreen, #green and #default tags from the supplied string"""
      return text.replace('#lightgreen', '').replace('#green', '').replace('#default', '')

bombtime = BombTimerManager()


###


dict_options   = {'cheapbombtimer_bps':[83, 'Bomb beeps per second (x 10)'], 'cheapbombtimer_sound_percent':[50, 'Percent of the map around the bomb the voice can be heard'], 'cheapbombtimer_announce_chat':[0, '0 = no change, 1 = announce bomb time in chat area'], 'cheapbombtimer_announce_center':[1, '0 = no change, 1 = announce bomb time in center text'], 'cheapbombtimer_announce_hud':[0, '0 = no change, 1 = announce bomb time in a HUD hint'], 'cheapbombtimer_bps_skipcheck':[0, '0 = no change, 1 = CheapBombTimer will not check the BPS value if the bomb explodes (only applies to current round)']}
func_lang_text = lambda x, y, z: 'CheapBombTimer: cheapbombtimer_languages.ini not found'

def load():
   """
   Checks to make sure critical server variables are created by the config
   """
   global func_lang_text

   for str_option in dict_options:
      es.ServerVar(str_option, dict_options[str_option][0], dict_options[str_option][1])
   if os.path.isfile(es.getAddonPath('cheapbombtimer') + '/cheapbombtimer.cfg'):
      es.server.cmd('es_xmexec ../addons/eventscripts/cheapbombtimer/cheapbombtimer.cfg')
   else:
      es.dbgmsg(0, 'CheapBombTimer: Unable to load cheapbombtimer.cfg! Please ensure it is in the ./cheapbombtimer/ directory.')

   str_lang_path = es.getAddonPath('cheapbombtimer') + '/cheapbombtimer_languages.ini'
   if os.path.isfile(str_lang_path):
      func_lang_text = langlib.Strings(str_lang_path)
   else:
      es.dbgmsg(0, 'CheapBombTimer: Unable to load cheapbombtimer_languages.ini! Please ensure it is in the ./cheapbombtimer/ directory.')

   bombsound.markDownloads()


def es_map_start(event_var):
   bombsound.markDownloads()


def round_start(event_var):
   bombtime.resetTicks()


def bomb_planted(event_var):
   bombtime.bombPlanted()


def bomb_beep(event_var):
   bombtime.tick(int(event_var['entindex']))


def bomb_exploded(event_var):
   bombtime.bombExploded()


def unload():
   bombtime.resetTicks()

   for str_option in dict_options:
      es.ServerVar(str_option).set(0)


###


def addsound_cmd():
   """
   cheapbombtimer_addsound <"sound id"> <time> <"sound relative to ./sound/">
   """
   if es.getargc() == 4:
      bombsound.addSound(es.getargv(1), round(float(es.getargv(2)), 2), es.getargv(3))

   else:
      es.dbgmsg(0, 'Syntax: cheapbombtimer_addsound <\"sound id\"> <time> <\"sound relative to ./sound/\">')

if not es.exists('command', 'cheapbombtimer_addsound'):
   es.regcmd('cheapbombtimer_addsound', 'cheapbombtimer/addsound_cmd', 'cheapbombtimer_addsound <\"sound id\"> <time> <\"sound relative to ./sound/\">\nAdds bomb timer sound based on sound id and bomb time')


def removesound_cmd():
   """
   cheapbombtimer_removesound <"sound id">
   """
   if es.getargc() == 2:
      bombsound.removeSound(es.getargv(1))

   else:
      es.dbgmsg(0, 'Syntax: cheapbombtimer_removesound <\"sound id\">')

if not es.exists('command', 'cheapbombtimer_removesound'):
   es.regcmd('cheapbombtimer_removesound', 'cheapbombtimer/removesound_cmd', 'cheapbombtimer_removesound <\"sound id\">\nRemoves bomb timer sound based on sound id')


def addremovetext_cmd():
   """
   cheapbombtimer_addtext <time 1> [time 2] ... [time n]
   cheapbombtimer_removetext <time 1> [time 2] ... [time n]
   """
   arg_count = es.getargc()
   if arg_count > 1:
      times = map(lambda x: int(es.getargv(x)), range(1, arg_count))

      if es.getargv(0).lower() == 'cheapbombtimer_addtext':
         bombtext.addAnnouncements(times)

      else:
         bombtext.removeAnnouncements(times)

   else:
      es.dbgmsg(0, 'Syntax: %s <time 1> [time 2] ... [time n]' % es.getargv(0).lower())

if not es.exists('command', 'cheapbombtimer_addtext'):
   es.regcmd('cheapbombtimer_addtext', 'cheapbombtimer/addremovetext_cmd', 'cheapbombtimer_addtext <time 1> [time 2] ... [time n]\nAdds bomb time text announcements')

if not es.exists('command', 'cheapbombtimer_removetext'):
   es.regcmd('cheapbombtimer_removetext', 'cheapbombtimer/addremovetext_cmd', 'cheapbombtimer_removetext <time 1> [time 2] ... [time n]\nRemoves bomb time text announcements')